/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package vcs.commands;

import java.util.*;

import org.openide.*;
import org.openide.util.*;

//import org.netbeans.modules.vcscore.cmdline.*;
//import org.netbeans.modules.vcscore.cmdline.exec.*;
import org.netbeans.modules.vcscore.VcsFileSystem;
import org.netbeans.modules.vcscore.Variables;
import org.netbeans.modules.vcscore.VcsConfigVariable;
import org.netbeans.modules.vcscore.commands.*;
import org.netbeans.modules.vcscore.util.*;
//import org.netbeans.modules.vcscore.cmdline.VcsVariableSelector;
import org.netbeans.modules.vcscore.cmdline.VcsAdditionalCommand;

/**
 *
 * @author  Martin Entlicher
 */
public class CvsModuleSelector implements VcsAdditionalCommand, CommandDataOutputListener {
    private Debug E = new Debug("CvsModuleSelector", true); // NOI18N
    private Debug D = E;

    private Hashtable vars;
    private CommandOutputListener stdoutNRListener;
    private CommandOutputListener stderrNRListener;
    private String dataRegex = "^(.*)$";
    private StringBuffer outputBuffer = new StringBuffer();
    private volatile boolean cmdSuccess = false;
    private volatile boolean dlgSuccess = false;
    private volatile boolean dlgFinished = false;
    private boolean isCommand = false;
    
    private static transient String lastPrepared1 = null;
    private static transient String lastPrepared2 = null;
    private static transient Vector lastModules = null;
    
    //private volatile boolean modulesStatGetStatus = false;
    //private volatile boolean modulesAllGetStatus = false;
    private volatile CompleteStatusOutput completeOutput = null;
    private VcsFileSystem fileSystem = null;

    /** Creates new CvsModuleSelector */
    public CvsModuleSelector() {
        isCommand = false;
    }
    
    public void setFileSystem(VcsFileSystem fileSystem) {
        this.fileSystem = fileSystem;
    }

    /**
     * Find out whether we need to run the command which get module info.
     * If we have this information from previous call, we don't have to call it again.
     * @param args the command we want to run.
     * @return true or false
     */
    private boolean needToRunCommand(String[] args) {
        if (args.length != 2) {
            return true;
        }
        if (lastPrepared1 == null || lastPrepared2 == null || lastModules == null) return true;
        String cmd = args[0]; //VcsUtilities.array2string(args);
        String prepared = Variables.expand(vars, cmd, true);
        if (!prepared.equals(lastPrepared1)) return true;
        cmd = args[1]; //VcsUtilities.array2string(args);
        prepared = Variables.expand(vars, cmd, true);
        return !prepared.equals(lastPrepared2);
    }
    
    /** this method is used to start the selector from the CvsAction popup menu.
     *  This method is used to execute the command.
     * @param vars the variables that can be passed to the command
     * @param args the command line parametres passed to it in properties
     * @param stdoutNRListener listener of the standard output of the command
     * @param stderrNRListener listener of the error output of the command
     * @param stdoutListener listener of the standard output of the command which
     *                      satisfies regex <CODE>dataRegex</CODE>
     * @param dataRegex the regular expression for parsing the standard output
     * @param stderrListener listener of the error output of the command which
     *                      satisfies regex <CODE>errorRegex</CODE>
     * @param errorRegex the regular expression for parsing the error output
     * @return true if the command was succesfull
     *        false if some error occured.
     */
    public boolean exec(Hashtable vars,String[] args,
                        CommandOutputListener stdoutNRListener,
                        CommandOutputListener stderrNRListener,
                        CommandDataOutputListener stdoutListener, String dataRegex,
                        CommandDataOutputListener stderrListener, String errorRegex) {
        isCommand = true;
        String[] returns = execSel(vars, "MODULE", args, stdoutNRListener, stderrNRListener);
        String[] toReturn = new String[1];
        if (returns != null) {
          if (returns[0] != null) { 
            toReturn[0] = VcsUtilities.arrayToQuotedString(returns, false);  //TODO for cygwin =true
            stdoutListener.outputData(toReturn); 
            return true;
          }  
        }
        //toReturn[0] = ".";
        //stdoutListener.match(toReturn);
        return false;  
    }
    
    /**
     * This method is used to start the selector.
     * @param vars the VCS variables
     * @param variable the name of the selected variable
     * @param args the command line parametres
     * @param stdoutNRListener listener of the standard output of the command
     * @param stderrNRListener listener of the error output of the command
     * @return the selected values, empty string when the selection was canceled
     *         or null when an error occures.
     */
    public String[] execSel(Hashtable vars, String variable, String[] args,
                       CommandOutputListener stdoutNRListener,
                       CommandOutputListener stderrNRListener) {
        D.deb("exec for "+variable);
        this.vars = vars;
        this.stdoutNRListener = stdoutNRListener;
        this.stderrNRListener = stderrNRListener;
        /*
        if (!runCommand(args)) return null;
        String[] modules = getModules();
        if (modules == null || modules.length <= 0) {
          javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
              NotifyDescriptor nd = new NotifyDescriptor.Message (org.openide.util.NbBundle.getBundle(CvsModuleSelectorDialog.class).getString("CvsModuleSelectorDialog.NoModules"));
              TopManager.getDefault ().notify (nd);
            }
          });
          return null;
    }
        */
        javax.swing.JFrame dlgFrame = new javax.swing.JFrame ();
        VcsUtilities.centerWindow(dlgFrame);
        CvsModuleSelectorDialog dlg = new CvsModuleSelectorDialog(dlgFrame, true);
        VcsUtilities.centerWindow(dlg);
        dlg.calledAsCommand(isCommand);  // if run as command -> is called from CvsAction - disable some stuff

        final String[] fargs = args;
        final CvsModuleSelectorDialog fdlg = dlg;
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                javax.swing.SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        new Thread("ModuleSelector-Command") {
                            public void run() {
                                Vector modules = null;
                                if (needToRunCommand(fargs)) {
                                    fdlg.waitingForModules();
                                    cmdSuccess = runCommand(fargs);
                                    if (cmdSuccess) modules = getModules();
                                } else {
                                    cmdSuccess = true;
                                    modules = lastModules;
                                }
                                fdlg.setModules(modules, cmdSuccess);
                            }
                        }.start();
                    }
                });
            }
        });
        //dlg.setModules(modules);
        Thread showThread = new Thread() {
            public void run() {
                dlgSuccess = fdlg.showDialog();
                dlgFinished = true;
            }
        };
        javax.swing.SwingUtilities.invokeLater(showThread);
        try {
            while(!dlgFinished) {
                Thread.sleep(200);
            }
            D.deb("showThread is alive = "+showThread.isAlive()+", joining him");
            showThread.join();
        } catch (InterruptedException e) {
            // Interrupted
            dlgSuccess = false;
        }
        D.deb("dlgSuccess = "+dlgSuccess+", cmdSuccess = "+cmdSuccess);
        if (dlgSuccess && cmdSuccess) {
            return dlg.getSelection();
        } else {
            if (dlgSuccess) return null;
            else {
              String[] toReturn = {""};
              return toReturn;                        
            }    
        }
    }

    private boolean runCommand(String[] args) {
        if (args.length != 2) {
            if (stderrNRListener != null) stderrNRListener.outputLine("Bad number of arguments. "+
                                                                      "Expecting two arguments: cvs co -s AND cvs co -c");
            return false;
        }
        //String cmd = args[0]; //VcsUtilities.array2string(args);
        VcsCommand cmd = fileSystem.getCommand(args[0]);
        //final String prepared1 = Variables.expand(vars, cmd, true);
        //D.deb("prepared1: "+prepared1); // NOI18N
        //final ExternalCommand ec1 = new ExternalCommand(prepared1);
        //ec.setTimeout(((Long) vars.get("TIMEOUT")).longValue()); // NOI18N
        Hashtable varsCopy1 = new Hashtable(vars);
        VcsCommandExecutor vce1 = fileSystem.getVcsFactory().getCommandExecutor(cmd, varsCopy1);
        vce1.addDataOutputListener(this);
        /*
        if (stdoutNRListener != null) ec1.addStdoutNoRegexListener(stdoutNRListener);
        if (stderrNRListener != null) ec1.addStderrNoRegexListener(stderrNRListener);
        if (stderrNRListener != null) stderrNRListener.match(NbBundle.getBundle
                    ("vcs.commands.Bundle").getString ("MSG_VariableSelector")+
                    ": "+prepared1); // NOI18N
        try{
            ec1.addStdoutRegexListener(this, dataRegex);
        } catch (BadRegexException e) {
            if (stderrNRListener != null) {
                stderrNRListener.match("Bad data regex "+dataRegex); // NOI18N
            }
            return false;
        }
         */
        cmd = fileSystem.getCommand(args[1]);
        //final String prepared2 = Variables.expand(vars, cmd, true);
        //final ExternalCommand ec2 = new ExternalCommand(prepared2);
        Hashtable varsCopy2 = new Hashtable(vars);
        VcsCommandExecutor vce2 = fileSystem.getVcsFactory().getCommandExecutor(cmd, varsCopy2);
        completeOutput = new CompleteStatusOutput();
        vce2.addDataOutputListener(completeOutput);
        /*
        if (stdoutNRListener != null) ec2.addStdoutNoRegexListener(stdoutNRListener);
        if (stderrNRListener != null) ec2.addStderrNoRegexListener(stderrNRListener);
        if (stderrNRListener != null) stderrNRListener.match(NbBundle.getBundle
                    ("vcs.commands.Bundle").getString ("MSG_VariableSelector")+
                    ": "+prepared2); // NOI18N
        try{
            ec2.addStdoutRegexListener(completeOutput, dataRegex);
        } catch (BadRegexException e) {
            if (stderrNRListener != null) {
                stderrNRListener.match("Bad data regex "+dataRegex); // NOI18N
            }
            return false;
        }
         */
        CommandsPool pool = fileSystem.getCommandsPool();
        pool.startExecutor(vce1);
        pool.startExecutor(vce2);
        pool.waitToFinish(vce1);
        pool.waitToFinish(vce2);
        boolean modulesStatGetStatus = vce1.getExitStatus() == VcsCommandExecutor.SUCCEEDED;
        boolean modulesAllGetStatus = vce2.getExitStatus() == VcsCommandExecutor.SUCCEEDED;
        /*
        Thread t1 = new Thread("CVS-ModuleSelector1") {
            public void run() {
                if (ec1.exec() != ExternalCommand.SUCCESS) {
                    //E.err("exec failed "+ec1.getExitStatus()); // NOI18N
                    modulesStatGetStatus = false;
                } else {
                    lastPrepared1 = prepared1;
                    modulesStatGetStatus = true;
                }
            }
        };
        Thread t2 = new Thread("CVS-ModuleSelector2") {
            public void run() {
                if (ec2.exec() != ExternalCommand.SUCCESS) {
                    //E.err("exec failed "+ec1.getExitStatus()); // NOI18N
                    modulesAllGetStatus = false;
                } else {
                    lastPrepared2 = prepared2;
                    modulesAllGetStatus = true;
                }
            }
        };
        t1.start();
        t2.start();
        try {
            t1.join();
            t2.join();
        } catch (InterruptedException exc) {
            t1.interrupt();
            t2.interrupt();
        }
         */
        return modulesStatGetStatus && modulesAllGetStatus;
    }

    private Vector getModules() {
        Vector modules = new Vector();
        String output = outputBuffer.toString();
       //MK rewrite to add status of the modules..  
        String line;
        int index = 0;
        StringBuffer module;
        int addedCount = 0;
        StringTokenizer token = new StringTokenizer(output, "\n", false);
        while (token.hasMoreTokens()) {
            Vector row = new Vector();
            line = token.nextToken();
            if (line.startsWith(" ") || line.startsWith("\t")) continue; // it's not a line with module name
            line = line.replace('\t',' ');
            StringTokenizer lineTok = new StringTokenizer(line, " ", false);
            if (lineTok.countTokens() < 2) continue;
            String prvni = lineTok.nextToken();
            row.add(prvni);
            row.add(lineTok.nextToken());
            modules.add(row);
        }
        completeOutput.insertDifferent(modules);
        return modules;  
    }

    public void outputData(String[] elements) {
        //D.deb("match: "+elements[0]);
        if (elements[0].length() > 0 && elements[0].charAt(0) != '#')
            outputBuffer.append(elements[0]+"\n");
    }

    private final class CompleteStatusOutput implements CommandDataOutputListener {
        private LinkedList statuses;
        
        public CompleteStatusOutput() {
            statuses = new LinkedList();
        }
        
        public void removeFromList(String name) {
            statuses.remove(name);
        }
        
        public void  insertDifferent(Vector modules) {
            Iterator parIter = modules.iterator();
            while(parIter.hasNext()) { // first remove all that are in the list with statuses
                Vector row = (Vector)parIter.next();
                removeFromList((String)row.get(0));
            }
            Iterator it = statuses.iterator();
            while (it.hasNext()) {
                Vector row = new Vector();
                String str = (String)it.next();
                //D("in -c list:" + str);
                row.add(str);
                row.add("");
                modules.add(row);
            }
        }
        
        /** method from the NoRegExpListener.
         * reads al output of the checkout -c command.
         * @param elements lines.
         */
        public void outputData(String[] elements) {
            if (elements[0].length() > 0 && elements[0].charAt(0) != '#') {
                String start = elements[0];
                if (start == null) return;
                if (!start.startsWith(" ")) {
                    int index = start.indexOf(' ');
                    if (index > 0) {
                        String name = start.substring(0,index);
                        statuses.add(name);
                    }
                }
            }
        }
        
    }

}
