/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package vcs.commands;

import java.io.*;
import java.util.*;

import org.openide.filesystems.FileObject;
//import org.openide.util.enum.AlterEnumeration;

import org.netbeans.modules.vcscore.Variables;
import org.netbeans.modules.vcscore.VcsFileSystem;
import org.netbeans.modules.vcscore.VcsAction;
import org.netbeans.modules.vcscore.commands.*;
import org.netbeans.modules.vcscore.util.*;
//import org.netbeans.modules.vcscore.cmdline.exec.*;
import org.netbeans.modules.vcscore.cmdline.VcsAdditionalCommand;

import org.netbeans.modules.vcs.advanced.CommandLineVcsFileSystem;

/**
 * This class works as a wrapper for recursive commands which needs to perform
 * some tasks in each subdirectory of a current directory.
 * Due to problems with filesystem.attributes in PVCS, the command is called explicitly
 * for EACH file in the directory excluding filesystem.attributes
 * @author  Martin Entlicher
 */
public class RecursiveFolderCommand extends Object implements VcsAdditionalCommand {

    private static final String IGNORE_FILE = "filesystem.attributes";
    
    private VcsFileSystem fileSystem = null;
    private CommandOutputListener stdoutNRListener;
    private CommandOutputListener stderrNRListener;
    private CommandDataOutputListener stdoutListener;
    private CommandDataOutputListener stderrListener;
    private String dataRegex;
    private String errorRegex;
    private Collection filteredFilesOut = null;
    private boolean filteredFilesOutCaseInsensitive = false;
    
    private String rootDir = "";

    /** Creates new RecursiveFolderCommand */
    public RecursiveFolderCommand() {
    }
    
    public void setFileSystem(VcsFileSystem fileSystem) {
        this.fileSystem = fileSystem;
    }
    
    /*
    private String getInitDir(Hashtable vars) {
        String dir;
        this.rootDir = (String) vars.get("ROOTDIR"); // NOI18N
        if (rootDir == null) {
            rootDir = "."; // NOI18N
        }
        dir = (String) vars.get("PATH"); // NOI18N
        if (dir == null) {
            dir = ""; // NOI18N
        }
        String module = (String) vars.get("MODULE"); // NOI18N
        //D.deb("rootDir = "+rootDir+", module = "+module+", dir = "+dir); // NOI18N
        if (dir.equals("")) { // NOI18N
            dir = rootDir;
            if (module != null && module.length() > 0) dir += File.separator + module;
        } else {
            if (module == null)
                dir = rootDir + File.separator + dir;
            else
                dir = rootDir + File.separator + module + File.separator + dir;
        }
        if (dir.charAt(dir.length() - 1) == File.separatorChar) {
            dir = dir.substring(0, dir.length() - 1);
        }
        //D.deb("dir="+dir); // NOI18N
        return dir;
    }
     */
    
    /**
     * Get recursively all subdirectories inside the directory.
     * @return the array of absolute pathnames of all subdirectories.
     *
    private File[] getFiles(String dir) {
        File dirFile = new File(dir);
        if (!dirFile.exists() || !dirFile.isDirectory()) return null;
        File[] subfiles = dirFile.listFiles(new FileFilter() {
            public boolean accept(File file) {
                return file.isDirectory() && !filteredFilesOut.contains((filteredFilesOutCaseInsensitive)
                                                                        ? file.getName().toUpperCase()
                                                                        : file.getName());
            }
        });
        return subfiles;
    }
    
    private String[] getChildren(String folder) {
        return fileSystem.children(folder);
    }
     */

    /*
    private CommandRunnable getExternalCommand(Hashtable vars, String cmd) {
        String prepared = Variables.expand(vars, cmd, true);
        //System.out.println("prepared = "+prepared);
        final ExternalCommand ec = new ExternalCommand(prepared);
        String input = (String) vars.get("INPUT");
        if (input == null) input = "";
        long timeout = ((Long) vars.get("TIMEOUT")).longValue();
        ec.setTimeout(timeout);
        ec.setInput(input);
        if (stdoutNRListener != null) ec.addStdoutNoRegexListener(stdoutNRListener);
        if (stderrNRListener != null) ec.addStderrNoRegexListener(stderrNRListener);
        if (stdoutListener != null) {
            try {
                ec.addStdoutRegexListener(stdoutListener, dataRegex);
            } catch(BadRegexException exc) {
                org.openide.TopManager.getDefault().notifyException(exc);
            }
        }
        if (stderrListener != null) {
            try {
                ec.addStderrRegexListener(stderrListener, errorRegex);
            } catch(BadRegexException exc) {
                org.openide.TopManager.getDefault().notifyException(exc);
            }
        }
        CommandRunnable command = new CommandRunnable(ec);
        return command;
    }
     */
    
    /*
    private void setFilteredFilesOut(final Hashtable vars) {
        String filteredFiles = (String) vars.get(CommandLineVcsFileSystem.VAR_LOCAL_FILES_FILTERED_OUT);
        if (filteredFiles != null) {
            String[] files = VcsUtilities.getQuotedStrings(filteredFiles);
            filteredFilesOut = new HashSet(Arrays.asList(files));
            filteredFilesOutCaseInsensitive = VcsFileSystem.VAR_FALSE.equals(vars.get(CommandLineVcsFileSystem.VAR_LOCAL_FILES_FILTERED_OUT_CASE_SENSITIVE));
            if (filteredFilesOutCaseInsensitive) {
                Enumeration upperCaseFiles = new AlterEnumeration(Collections.enumeration(filteredFilesOut)) {
                    protected Object alter(Object o) {
                        return ((String) o).toUpperCase();
                    }
                };
                filteredFilesOut = new HashSet();
                while(upperCaseFiles.hasMoreElements()) filteredFilesOut.add(upperCaseFiles.nextElement());
            }
        }
        if (filteredFilesOut == null) {
            filteredFilesOut = Collections.singleton(IGNORE_FILE);
        } else {
            filteredFilesOut.add((filteredFilesOutCaseInsensitive) ? IGNORE_FILE.toUpperCase() : IGNORE_FILE);
        }
    }
     */
    
    private boolean runCommandRecursively(final FileObject dir, final Hashtable vars, final VcsCommand cmd) {
        /*
        String relWork;
        if (dir.length() > rootDir.length()) {
            relWork = dir.substring(rootDir.length() + 1); // the folder relative to the root directory.
        } else {
            relWork = ""; // the folder relative to the root directory.
        }
        while(relWork.length() > 0 && relWork.charAt(0) == File.separatorChar) relWork = relWork.substring(1);
        vars.put("RELWORK", relWork); // the folder relative to the root directory.
        //System.out.println("RELWORK = '"+relWork+"'");
        File dirFile = new File(dir);
        final File[] subfiles = dirFile.listFiles(new FileFilter() {
            public boolean accept(File file) {
                return (file.isFile() && !filteredFilesOut.contains((filteredFilesOutCaseInsensitive)
                                                                    ? file.getName().toUpperCase()
                                                                    : file.getName()));
            }
        });
         */
        FileObject[] children = dir.getChildren();
        ArrayList childrenFiles = new ArrayList();
        ArrayList childrenFolders = new ArrayList();
        for (int i = 0; i < children.length; i++) {
            FileObject fo = children[i];
            if (fo.isFolder()) childrenFolders.add(fo);
            else childrenFiles.add(fo);
        }
        Table files = new Table();
        boolean processAll = VcsCommandIO.getBooleanPropertyAssumeDefault(cmd, VcsCommand.PROPERTY_PROCESS_ALL_FILES) || fileSystem.isProcessUnimportantFiles();
        VcsAction.addImportantFiles(childrenFiles, files, processAll, fileSystem);
        VcsCommandExecutor[] commands = VcsAction.doCommand(files, cmd, null, fileSystem,
                                                            stdoutNRListener, stderrNRListener, stdoutListener, stderrListener);
        boolean success = true;
        for (Iterator it = childrenFolders.iterator(); it.hasNext(); ) {
            FileObject fo = (FileObject) it.next();
            Hashtable newVars = new Hashtable(vars);
            boolean recSuccess = runCommandRecursively(fo, newVars, cmd);
            success = success && recSuccess;
        }
        for (int i = 0; i < commands.length; i++) {
            VcsCommandExecutor vce = commands[i];
            fileSystem.getCommandsPool().waitToFinish(vce);
            success = success && (vce.getExitStatus() == VcsCommandExecutor.SUCCEEDED);
        }
        return success;
    }

    /**
     * This method is used to execute the command.
     * The provided command is executed recursively for all subfolders.
     * The command is executed only on non-empty folders.
     *
     * @param vars the variables that can be passed to the command
     * @param args the command line parametres passed to it in properties
     * @param stdoutNRListener listener of the standard output of the command
     * @param stderrNRListener listener of the error output of the command
     * @param stdoutListener listener of the standard output of the command which
     *                      satisfies regex <CODE>dataRegex</CODE>
     * @param dataRegex the regular expression for parsing the standard output
     * @param stderrListener listener of the error output of the command which
     *                      satisfies regex <CODE>errorRegex</CODE>
     * @param errorRegex the regular expression for parsing the error output
     * @return true if the command was succesfull
     *        false if some error occured.
     */
    public boolean exec(Hashtable vars, String[] args, CommandOutputListener stdoutNRListener,
                        CommandOutputListener stderrNRListener,
                        CommandDataOutputListener stdoutListener, String dataRegex,
                        CommandDataOutputListener stderrListener, String errorRegex) {
        this.stdoutNRListener = stdoutNRListener;
        this.stderrNRListener = stderrNRListener;
        this.stdoutListener = stdoutListener;
        this.stderrListener = stderrListener;
        this.dataRegex = dataRegex;
        this.errorRegex = errorRegex;
        //String dir = getInitDir(vars);
        //String cmd = VcsUtilities.array2string(args);
        String cmd = args[0];
        //setFilteredFilesOut(vars);
        FileObject dir = fileSystem.findResource((String) vars.get("DIR"));
        return runCommandRecursively(dir, vars, fileSystem.getCommand(cmd));

    }
    
}
